# -*- encoding: utf-8 -*-
"""
hio.core.memo Package
"""


from .memoing import (Versionage, Sizage, GramDex, SGDex,
                      openMemoer, Memoer, MemoerDoer,
                      openTM, TymeeMemoer, TymeeMemoerDoer)

